@@ -2,7 +2,7 @@
  * refind/install.c
  * Functions related to installation of rEFInd and management of EFI boot order
  *
- * Copyright (c) 2020-2023 by Roderick W. Smith
+ * Copyright (c) 2020-2024 by Roderick W. Smith
  *
  * Distributed under the terms of the GNU General Public License (GPL)
  * version 3 (GPLv3), a copy of which must be distributed with this source
@@ -55,7 +55,7 @@ static ESP_LIST * FindAllESPs(VOID) {
     UINTN VolumeIndex;
     EFI_GUID ESPGuid = ESP_GUID_VALUE;
 
-    LOG(2, LOG_LINE_NORMAL, L"Searching for ESPs");
+    LOG(2, LOG_LINE_NORMAL, L"Checking %ld volumes to find ESP(s)", VolumesCount);
     for (VolumeIndex = 0; VolumeIndex < VolumesCount; VolumeIndex++) {
         if (Volumes[VolumeIndex]->DiskKind == DISK_KIND_INTERNAL &&
             GuidsAreEqual(&(Volumes[VolumeIndex]->PartTypeGuid), &ESPGuid) &&
@@ -156,6 +156,12 @@ static REFIT_VOLUME *PickOneESP(ESP_LIST *AllESPs) {
  *
  ***********************/
 
+// Rename OldName to NewName in BaseDir.
+// CAUTION: On a Raspberry Pi running Tow-Boot (aka Das U-Boot) 8225.40, the
+// attempt to rename the file (specific point called out below) can fail,
+// which results in a subsequent crash if an attempt is made to create a new
+// file using the old (supposedly changed) name. Checking the return value
+// is therefore important for this call!
 static EFI_STATUS RenameFile(IN EFI_FILE_PROTOCOL *BaseDir, CHAR16 *OldName, CHAR16 *NewName) {
     EFI_STATUS        Status;
     EFI_FILE_PROTOCOL *FilePtr = NULL; // DO NOT FREE!
@@ -176,9 +182,10 @@ static EFI_STATUS RenameFile(IN EFI_FILE_PROTOCOL *BaseDir, CHAR16 *OldName, CHA
         NewInfoSize = sizeof(EFI_FILE_INFO) + StrSize(NewName);
         NewInfo = (EFI_FILE_INFO *) AllocateZeroPool(NewInfoSize);
         if (NewInfo != NULL) {
-            CopyMem(NewInfo, Buffer, sizeof(EFI_FILE_INFO));
+            MyCopyMem(NewInfo, Buffer, sizeof(EFI_FILE_INFO));
             NewInfo->FileName[0] = 0;
             StrCat(NewInfo->FileName, NewName);
+            // Note: The below call is where Tow-Boot can error out.
             Status = refit_call4_wrapper(BaseDir->SetInfo,
                                          FilePtr,
                                          &gEfiFileInfoGuid,
@@ -416,7 +423,11 @@ static EFI_STATUS CopyFiles(IN EFI_FILE_PROTOCOL *TargetDir) {
         //    copy to refind.conf-sample if refind.conf is present.
         ConfFile = PoolPrint(L"%s\\refind.conf-sample", SourceDir);
         if (FileExists(SourceVolume->RootDir, ConfFile)) {
-            StrCpy(SourceFile, ConfFile);
+            // The following line originally used StrCpy(), but that had a bug
+            // because it didn't allocate SourceFile; but even after fixing
+            // that, one particular UEFI froze when freeing SourceFile later.
+            // Using PoolPrint() rather than StrCpy() works around this issue.
+            SourceFile = PoolPrint(L"%s", ConfFile);
         } else {
             SourceFile = PoolPrint(L"%s\\refind.conf", SourceDir);
         }
@@ -596,7 +607,7 @@ static EFI_STATUS ConstructBootEntry(EFI_HANDLE *TargetVolume,
         Working += sizeof (UINT16);
         StrCpy((CHAR16 *)Working, Label);
         Working += StrSize(Label);
-        CopyMem(Working, DevicePath, DevPathSize);
+        MyCopyMem(Working, DevicePath, DevPathSize);
         // If support for arguments is required in the future, uncomment
         // the below two lines and adjust Size computation above appropriately.
         // Working += DevPathSize;
@@ -687,7 +698,7 @@ VOID InstallRefind(VOID) {
     REFIT_VOLUME  *SelectedESP; // Do not free
     UINTN         Status;
 
-    LOG(1, LOG_LINE_NORMAL, L"Installing rEFInd to an ESP");
+    LOG(1, LOG_LINE_THIN_SEP, L"Installing rEFInd to an ESP");
     AllESPs = FindAllESPs();
     SelectedESP = PickOneESP(AllESPs);
     if (SelectedESP) {
@@ -701,6 +712,7 @@ VOID InstallRefind(VOID) {
             DisplaySimpleMessage(L"Warning", L"Problems encountered during installation");
         } // if/else
     } // if
+    LOG(1, LOG_LINE_THIN_SEP, L"Done installing rEFInd to an ESP");
 } // VOID InstallRefind()
 
 /***********************
@@ -735,7 +747,7 @@ BOOT_ENTRY_LIST * FindBootOrderEntries(VOID) {
                 L->BootEntry.Size = (UINT16) Contents[2];
                 L->BootEntry.Label = StrDuplicate((CHAR16*) &(Contents[3]));
                 L->BootEntry.DevPath = AllocatePool(L->BootEntry.Size);
-                CopyMem(L->BootEntry.DevPath,
+                MyCopyMem(L->BootEntry.DevPath,
                         (EFI_DEVICE_PATH*) &Contents[3 + StrSize(L->BootEntry.Label)/2],
                         L->BootEntry.Size);
                 L->NextBootEntry = NULL;
